// CForth  project ************************************************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.h","cforth.c","cforth.h"
// and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32g431xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{
uint32_t k;
uint32_t recvword;
uint8_t sByte;

    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;  //necessary for what??
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIOAEN;  // Enable Port A clock
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIOBEN;  // Enable Port B clock
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIOCEN;  // Enable Port B clock
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIODEN;  // Enable Port B clock

    GPIOA->MODER = 0x00000000;      //primarily all input
    GPIOB->MODER = 0x00000000;      //primarily all input
    GPIOC->MODER = 0x00000000;      //primarily all input
    GPIOD->MODER = 0x00000000;      //primarily all input

    GPIOA->PUPDR = 0xA402AAAA;      //primarily all inputs pulldown,
                                    //except PA9,10,11,12,13,14
    GPIOB->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown
    GPIOC->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown
    GPIOD->PUPDR = 0x00000020;	//primarily PD2 pullDown (only existing one)

    SetSysClock();

//pin config repeated specifically:
	GPIOA->PUPDR &= ~(GPIO_PUPDR_PUPDR5);	//pull-down off, wasts current
    GPIOA->MODER &= ~(GPIO_MODER_MODE5);
    GPIOA->MODER |= GPIO_MODER_MODE5_0 ;  	//set PortA5 (user LED) as output
    GPIOA->BSRR =  GPIO_BSRR_BS_5;        	//PA5 (user LED) ON

    gSYSCOUNT = 50;
    gLEDCOUNT = 0;
    gDELAYCOUNT = 0;
    gTIMEOUT = 0;
    gProject = 0;
	gfRandom = 314159265;		//seed for xorshift32 algorithm
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;

    SysTick_Config(SystemCoreClock/1000);

    msDelay(100);

//start Independent Watchdog  **************
    IWDG->KR = 0x00005555;
    IWDG->RLR = 0x0000FFFF;
    msDelay(2);
    IWDG->PR = 0x00000002;      //resulting timeout ca. 1-2 seconds
    msDelay(2);
    IWDG->KR = 0x0000CCCC;      //start Independent Watchdog

//init I/O state  *********************
    popFlashPage23(); //only "gProject" is stored there
 //init communication  *********************
    initUsart2();						//always 115.2 kBaud
    msDelay(2);

    popFlashProject();                  //now load complete project

    while(1) {
		quit();
		query();
		if (gfTibMax == 0) continue;    //formally, in practice impossible
		eval();
    }
    msDelay(100);
    while(1);       //force cold reset by watchdog
}


//##########################################################################

// Misc Low Level Routines
//************************

void sendModel(void)  //  ****************************************************
//  **************************************************************************
{
const uint8_t typeStr [] = "CForth for STM32G431 Nucleo v";
const uint8_t copyStr [] = "  (c)2020-24 W.Schemmert";
const uint8_t clockStr [] = "  SystemClock=";

	sendTB(CRR);
	sendString(typeStr);
	sendTB(HARDWARE_VERSION + 0x30);
	sendTB('.');
	sendNS32(REVNUM);
	sendString(clockStr);
	sendNS32(SystemCoreClock / 1000000);
	sendTB('M');
	sendTB('H');
	sendTB('z');
	sendString(copyStr);
	sendTB(CRR);
}


void SetSysClock(void)  // *******************************************
//assumes "Vcore range 1", HSE clock 24.000 MHz,
//SystemClock=AHB=48MHz , APB1+2 = 48MHz
//Alternative: SystemClock=AHB=144MHz and APB1+2 = 72MHz
//adopted from "system_stm32f0xx.c" and modified according to STM32G431 FRefMan
//Reset SW[1:0], HPRE[3:0], PPRE1[2:0], PPRE2[2:0], MCOSEL[2:0] and MCOPRE[2:0]
// **************************************************************************
{
uint32_t StartUpCounter = 0, HSEStatus = 0;

#define HSE_STARTUP_TIMEOUT   0x50000 //Time out for HSE start up

    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;  //necessary for what??
//Vcore Range1 is default after Reset

// SYSCLK, HCLK, PCLK configuration ----------------------------------------
// Enable HSE
    RCC->CR |= RCC_CR_HSEON;

// Wait till HSE is ready or Timeout is reached
    do  {
      HSEStatus = RCC->CR & RCC_CR_HSERDY;
      StartUpCounter++;
     } while((HSEStatus == 0) && (StartUpCounter  < HSE_STARTUP_TIMEOUT));

#ifdef CPU_CLOCK144
     if (HSEStatus) {
// Enable Prefetch Buffer and set Flash Latency
        FLASH->ACR =
        FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |FLASH_ACR_ICEN |FLASH_ACR_LATENCY_7WS;

//System Clock shall be 144MHz
//  PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N
#define PLL_M      0        //24 MHz PLL input for project
#define PLL_N      12       //24*12=288MHz for project
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_Q      2        //divides PLL/6 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_R      0        //divides PLL/2 for 144MHz SYSCLK

 // HCLK = SYSCLK/
        RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 144MHz
 // PCLK2 = HCLK / 2
        RCC->CFGR |= RCC_CFGR_PPRE2_DIV2;   //set for project to 72MHz
 //PCLK1 = HCLK / 2
        RCC->CFGR |= RCC_CFGR_PPRE1_DIV2;   //set for project to 72MHz

//Configure the main PLL
//PLL input is 8MHz, so PLL_M is set correctly with next command:
//implies correct setting of PLL_R and PLL_Q, too
        RCC->PLLCFGR = PLL_N << 8;         //PLL_N = 36=0x12 = 001 0010 (bits 8-14)
        RCC->PLLCFGR |= PLL_Q << 21;        //PLL_Q = 2 (bits 21,22)
//    RCC->PLLCFGR |= PLL_R << 25;        //PLL_R = 00  (bits 25,26)

        RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
        RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1;            //select PLL_Q as USB clock
        RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN | RCC_PLLCFGR_PLLREN);
 //    }
#else
     if (HSEStatus) {
// Enable Prefetch Buffer and set Flash Latency
        FLASH->ACR =
        FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |FLASH_ACR_ICEN |FLASH_ACR_LATENCY_2WS;

//System Clock shall be 48MHz
//  PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N
#define PLL_M      0        //24 MHz PLL input for project
#define PLL_N      4       //24*4=96MHz for project
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_Q      0        //divides PLL/2 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_R      0        //divides PLL/2 for 48MHz SYSCLK

 // HCLK = SYSCLK/
    RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 48MHz
 // PCLK2 = HCLK / 2
    RCC->CFGR |= RCC_CFGR_PPRE2_DIV1;   //set for project to 48MHz
 //PCLK1 = HCLK / 2
    RCC->CFGR |= RCC_CFGR_PPRE1_DIV1;   //set for project to 48MHz

//Configure the main PLL
//PLL input is 8MHz, so PLL_M is set correctly with next command:
//implies correct setting of PLL_R and PLL_Q, too
    RCC->PLLCFGR = PLL_N << 8;         //PLL_N = 12=0x0C = 000 1100 (bits 8-14)
//    RCC->PLLCFGR |= PLL_Q << 21;        //PLL_Q = 00 (bits 21,22)
//    RCC->PLLCFGR |= PLL_R << 25;        //PLL_R = 00  (bits 25,26)

    RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
    RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1;            //select PLL_Q as USB clock
    RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN |  RCC_PLLCFGR_PLLREN);
#endif
//Select HSE as PLL source
    RCC->PLLCFGR |= RCC_PLLCFGR_PLLSRC_HSE; //set bits 0,1

// Enable the main PLL
    RCC->CR |= RCC_CR_PLLON;
// Wait till the main PLL is ready
    while((RCC->CR & RCC_CR_PLLRDY) == 0) ;

// Select PLL as system clock source
    RCC->CFGR |= RCC_CFGR_SW_PLL;

// Wait till PLL is used as system clock source
    while ((RCC->CFGR & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL) ;

    SystemCoreClockUpdate();//update in correspondence with parameters set above

    RCC->AHB1ENR |= RCC_AHB1ENR_FLASHEN; //necessary for Flash read/write action
  }
  else
  { // If HSE fails to start-up, the application will have wrong clock
    // configuration. User can add here some code to deal with this error
    //provisionally HSI16 is kept as SYSCLK
  }
}


void LED_Off(void)  // ******************************************************
// built as subroutine to supply service to "usb-fs-431.c" hardware independent
// **************************************************************************
{
    GPIOA->BSRR = GPIO_BSRR_BR_5; // clear bit PA5,LED OFF
    gLEDCOUNT = 4; // =ca.300 milliseconds
}


void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
	gDELAYCOUNT = count;
	while(gDELAYCOUNT)  servSystem();   //serve LED, watchdog etc.
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// *************************************************************************
{
uint32_t countdown;

#ifdef CPU_CLOCK144
	countdown = 87;
#else
	countdown = 51;
#endif

    while (countdown)   {
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
        countdown-- ;
    }
}


void servSystem(void)   // ************************************************
//not very time critical but permanently repeated functions for system control
//exclusively called in pullUsartRxBuf and pullUsbRxBuf
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 20 times/second
// ****************************************************************************
{
static uint32_t secCount;

	if (gSYSCOUNT)  return;			//countdown in SysTick_Handler()

//else if (gSYSCOUNT == 0)			//restart 20Hz service interval
    gSYSCOUNT = 50;
//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 1-2 seconds

//LED control
	if (gLEDCOUNT) gLEDCOUNT--;
	if (gLEDCOUNT==0)  GPIOA->BSRR =  GPIO_BSRR_BS_5; //PA5(User LED) ON again
}


void popFlashPage23(void)   // **********************************************
//read project independent system parameters
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t test, vid, pid;
uint32_t *sramPtr;

	emptyForth();           //init default system parameters

	flashPtr = (uint32_t *)PAGE23;
//vid and pid are dummies here
	test = ~(*flashPtr);//virgin test cannot be==0 if project was saved in Flash
	flashPtr++;
	vid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
	flashPtr++;
	pid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
	flashPtr++;

	if ((test | vid | vid)==0)  return; //defauls already init

	flashPtr++;					//compatibility dummy for gBaud

	gfNumBase =  ~(*flashPtr);
	flashPtr++;
	if ((gfNumBase!=10)&&(gfNumBase!=16)) gfNumBase = DEFAULT_NUMBASE;

	gProject =  ~(*flashPtr);   //number of project which is loaded by default
								//gProject=0 loads Empty Project
	if (gProject == 0)  emptyForth();
}


uint32_t pushFlashPage23(void)   //******************************************
//save project independent system parameters permanently
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t word;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
    else return 0;
 
//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 1-2 seconds

//erase Page23
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((23<<3) | FLASH_CR_EOPIE | FLASH_CR_ERRIE | FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write mirror into Page23:
//this microcontroller allows exclusively doubleword-wide write in flash
//flash least word is first word in readout
//i.e. first word in = first word read
//only write USB Vid/Pid, gBaud and gProject into page23
    flashPtr = (uint32_t *)PAGE23;

    FLASH->CR |= FLASH_CR_PG;
//dummy entries for compatibility of Flash with other projects
    *flashPtr = 0;                 //for flash virgin test
    flashPtr++;
    *flashPtr = ~(DEFAULT_VID & 0x0000FFFF);
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    *flashPtr = ~(DEFAULT_PID & 0x0000FFFF);
    flashPtr++;
    *flashPtr = ~DEFAULT_BAUD;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    *flashPtr = gfNumBase;
    flashPtr++;
    *flashPtr = ~gProject;

    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;

//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return  1;
}


void popFlashProject(void)   // *********************************************
//first read some project specific globals
//next read 2 pages to gfUserHeader[], then read 6 pages to gfTCode
// **************************************************************************
{
uint32_t k, m, n;
uint32_t * flashPtr;
uint32_t word, test;

//emptyForth() already updated in popFlashPage23() (default system parameters)

	if (gProject)  {
		flashPtr = (uint32_t *)(PAGE24 + 0x4000 * (gProject - 1));
	//first 32 bytes with system & peripheral config parameters
		test =  ~(*flashPtr);	//cannot be==0 if any project was saved in Flash
		flashPtr++;
		if (test == 0) {		//project Flash is virgin
			emptyForth();		//implies gProject=0 -> empty project
		}
		else  {
			gfNumBase= ~(*flashPtr);
			flashPtr++;
			if ((gfNumBase!=10)&&(gfNumBase!=16)) gfNumBase = DEFAULT_NUMBASE;
			gfTempNumBase = gfNumBase;

			gfAuto = ~(*flashPtr);
			flashPtr ++;

			k = ~(*flashPtr);
			flashPtr ++;
			gfSpiMode = (k >> 24) & 0x00000003;
			gfSpiBits = (k >> 16) & 0x00000003F;
			gfSpiClock = k & 0x00000FFFF;

			gfFreq12 = ~(*flashPtr);
			flashPtr ++;
			k = ~(*flashPtr);
			gfPWM1 = (int32_t)(k >> 16);
			if ((gfPWM1 == 0) || (gfPWM1 == 0x0000FFFF)) gfPWM1 = -1;
			gfPWM2 = (int32_t)(k & 0x0000FFFF); //Bugfix 16Dec21,before gfPWM1
			if ((gfPWM2 == 0) || (gfPWM2 == 0x0000FFFF)) gfPWM2 = -1;
			flashPtr ++;
			if (gfFreq12)  {
				subFREQ12();
				subPWM1();
				subPWM2();
			}
			gfFreq3 = ~(*flashPtr);
			flashPtr ++;
			k = ~(*flashPtr);
			gfPWM3 = (int32_t)k;
			if ((gfPWM3 == 0) || (gfPWM3 == 0x0000FFFF)) gfPWM3 = -1;
			flashPtr ++;
			if (gfFreq3)  {
				subFREQ3();
				subPWM3();
			}

			for (k = 0;k < MAXUSER/2;k++)  {  //reads 84*24 words=2016bytes +32=2048
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
			flashPtr = (uint32_t *)(PAGE25 + 0x4000 * (gProject - 1));
			for (k = (MAXUSER/2) +1;k < MAXUSER;k++) {  //reads 2044 bytes!
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
	
			flashPtr = (uint32_t *)(PAGE26 + 0x4000 * (gProject - 1));
			for (k = 0;k < MAXCODE;k++)  {      //reads 6 pages each 2048 bytes!
				gfTCode[k] = ~(*flashPtr);
				flashPtr ++;
			}
		}
	}
//from here if(gProject) or virgin or not:
	wait5us();
    op_ABORT();
	if (gfAuto)  {
//emergency brake to avoid AUTO deadlock
    	if (GPIOC->IDR & 0x2000) gfAuto = 0;	//if PC13=User Button is HIGH

	}									
	sendTB(CRR);
    op_QUEST();
    if (gfAuto) {
        gfParsed = gfAuto - 1;	//-1 to distinguish from empty AUTOEXE == 0
        interpretUserToken();
    }
}


uint32_t pushFlashProject(void)   //*****************************************
// **************************************************************************
{
uint32_t k, m, n, offset;
uint32_t * flashPtr;
uint32_t word[12];

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    if (gProject == 0) return 1;   //=default empty project, don't save

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 1-2 seconds

    sendTB('.');

    offset = (gProject - 1) * 8;     //project no.0 is empty, not treated here
                                    //project no.1 starts with page#24
//erase Page24 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((24 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//this microcontroller allows exclusively doubleword-wide write in flash
//flash least word is first word in readout
//i.e. first word in = first word read
//first write virgin test, relocalisation, gfNumBase and gfAuto into Page32

    flashPtr = (uint32_t *)(PAGE24 + 0x800 * offset); //0x4000*(project-1));
    FLASH->CR |= FLASH_CR_PG;
//first 32 bytes with system & peripheral config parameters
    *flashPtr = 0x00000000; //virgin flash test
    flashPtr++;
    *flashPtr = ~gfNumBase;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;

    *flashPtr = ~gfAuto;
    flashPtr++;
    *flashPtr = (~((gfSpiMode << 24) | (gfSpiBits << 16 ) |gfSpiClock));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~gfFreq12;
	flashPtr++;
    *flashPtr = ~((((uint32_t)gfPWM1 << 16) & 0xFFFF0000) | 
											((uint32_t)gfPWM2 & 0x0000FFFF));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~gfFreq3;
	flashPtr++;
    *flashPtr = ~((((uint32_t)gfPWM3) & 0xFFFF000));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

//write first half of gfUserHeader[] into Page24 + offset
    for (k = 0;k < MAXUSER/2; k+=2) {	//MAXUSER/2 = 84 ->2016bytes+32=2048
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;

        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;

        for (m = 0; m < 12; m+=2)  {
// this microcontroller allows exclusively doubleword-wide write in flash
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility with empty SRAM
            flashPtr++;
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page25 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((25 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second half of gfUserHeader[] into Page25 + offset
    flashPtr = (uint32_t *)(PAGE25 + 0x800 * offset); //0x4000 * (project - 1));

    FLASH->CR |= FLASH_CR_PG;
    for (k = (MAXUSER/2);k < MAXUSER; k+=2) {     //MAXUSER/2 = 84
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;
        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;
        for (m = 0; m < 12; m+=2)  {
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility with empty SRAM
            flashPtr++;
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page26 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((26 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write first part of gTCode into Page26 + offset
    flashPtr = (uint32_t *)(PAGE26 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
    for (k = 0; k < 0x200; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page27 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((27 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second part of gTCode into Page27 + offset

    flashPtr = (uint32_t *)(PAGE27 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
    for (k = 0x200; k < 0x400; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page28 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((28 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    FLASH->CR &= ~FLASH_CR_PER;

//write third part of gTCode into Page28
    flashPtr = (uint32_t *)(PAGE28 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
    for (k = 0x400; k < 0x600; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page29 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((29 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write fourth part of gTCode into Page29 + offset
    flashPtr = (uint32_t *)(PAGE29 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
    for (k = 0x600; k < 0x800; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page30 + ?ffset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((30 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
     FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;
    FLASH->CR &= ~FLASH_CR_PER;

//write fifth part of gTCode into Page30 + offset
    flashPtr = (uint32_t *)(PAGE30 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x800; k < 0xA00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page31 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((31 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write sixth part of gTCode into Page31 + offset
    flashPtr = (uint32_t *)(PAGE31 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0xA00; k < 0xC00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB('.');


//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return 1;
}

//###########################################################################

//USART Support
//*************

void initUsart2(void) // ****************************************************
// USART2-TxD is PortA,2
// USART2-RxD is PortA,3
// **************************************************************************
{
    RCC->APB1ENR1 |=  RCC_APB1ENR1_USART2EN ;   // USART2 clock enable

	GPIOA->MODER &= ~(GPIO_MODER_MODE2_0 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODE3_0 ) ;  // 10=Alternate function mode
	GPIOA->MODER |= (GPIO_MODER_MODE2_1 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODE3_1 ) ;  // 10=Alternate function mode
    //GPIOA->OTYPER is push-pull by default
    //GPIOA->OSPEEDR is 2 MHz by default
    //GPIOA->PUPDR is HiZ by default (except PA13,14,15)

//USART2 is AF7
    GPIOA->AFR[0] &= 0xFFFF00FF;
    GPIOA->AFR[0] |= 0x00007700;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;
#ifdef USART_TX_INTERRUPT
    gUSART_TXHEAD = gUSART_TXBUF;
    gUSART_TXTAIL = gUSART_TXBUF;
#endif

#ifdef CPU_CLOCK144
    USART2->BRR = 0x0271; //115200Baud: 625 at 72 MHz APB2 Clock
                          //only option when used together with onboard ST-Link
#else
    USART2->BRR = 0x01A0; //115200Baud: 416 at 48 MHz APB2 Clock
                          //only option when used together with onboard ST-Link
#endif
    USART2->CR2 = 0;			//default:1 Stopbit
    USART2->CR3 = 0;			//default:kein Handshake
    USART2->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |USART_CR1_TE | USART_CR1_RE;
                                //8Bit, no Parity, RX Int
    NVIC_EnableIRQ(USART2_IRQn);
        //effectively TX interrupt is enabled when bytes are put into the queue
}


void sendTB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
uint32_t fillstate;

  #ifdef USART_TX_INTERRUPT
    *gUSART_TXHEAD++ = data;
    if (gUSART_TXHEAD == gUSART_TXBUFTOP)  gUSART_TXHEAD = gUSART_TXBUF;
    USART2->CR1 |= USART_CR1_TXEIE;     //if new data renew interrupt enable
//added 02Aug2020
//calculate fill level of gUSB_TXBUF
//TAIL>HEAD? then add TXBUFLEN to get the absolute difference
	do {
		servSystem();   //serve LED, watchdog etc.
		if (gUSART_TXTAIL > gUSART_TXHEAD) {
		fillstate =(uint32_t)gUSART_TXHEAD + USART_TXBUFLEN -
													(uint32_t)gUSART_TXTAIL;
		}
		else fillstate = (uint32_t)gUSART_TXHEAD - (uint32_t)gUSART_TXTAIL ;
	} while (fillstate > (USART_TXBUFLEN-80));//buffer is emptied by interrupt
  #else
    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
  #endif

}


void sendNibble(uint8_t data) // ********************************************
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    sendTB(asciiNibble);
}

void sendString(const uint8_t * str) // *************************************
// **************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendTB(byte);
}


void sendX8(uint32_t data) // *********************************************
// send bits 0...7 in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    sendTB(SPACE);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendTB(SPACE);

}

void sendNX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble, noLead;
    noLead = 0;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    if (nibble)  {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
     if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
}


void send0X32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
//	if (data >= 10)  {			//modified 11Jan24
		sendTB('0');
		sendTB('x');
//	}
	sendNX32(data);
}


void sendS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendNS32(data);
    sendTB(SPACE);
}


 void sendLS32(uint32_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//leading SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendTB(SPACE);
    sendNS32(data);
}


void sendNS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//no leading and trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendTB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendTB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendTB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendTB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendTB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendTB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendTB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendTB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendTB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendTB((uint8_t)quotient + 0x30);
ONES:
    sendTB((uint8_t)data + 0x30) ;
}


uint32_t pullTermRxBuf(void) // *********************************************
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

    servSystem();   //serve LED, watchdog etc.

    if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
        recv= (uint32_t)((*gUSART_RXTAIL++ & 0x000000FF)|0x00008000);
        if (gUSART_RXTAIL==gUSART_RXBUFTOP) {
                gUSART_RXTAIL=gUSART_RXBUF;
        }

        GPIOA->BSRR = GPIO_BSRR_BR_5; // clear bit PA5,LED OFF
        gLEDCOUNT = 4; //=ca 300 milliseconds
        return recv;
    }
    else return 0;
}

// ##########################################################################
// Interrupt Handlers


void SysTick_Handler(void)  // ***********************************************
// ***************************************************************************
{
	if (gSYSCOUNT)  gSYSCOUNT--; //controls servSystem()
    if (gDELAYCOUNT) gDELAYCOUNT--;
    if (gTIMEOUT) gTIMEOUT--;  			//to be used in any subroutine
    if (gfMSCOUNT) gfMSCOUNT --;		//used by op_MS,op_DS
    if (gfTIXCOUNT) gfTIXCOUNT --;		//used by op__TIX, op_WAIT
	if (gfSTEPCOUNT1) gfSTEPCOUNT1--;	//stepper timing
	if (gfSTEPCOUNT2) gfSTEPCOUNT2--;
	if (gfSTEPCOUNT3) gfSTEPCOUNT3--;
	gfRandom ^= gfRandom << 13; //xorshift32 algorithm/white noise
	gfRandom ^= gfRandom >> 17;
	gfRandom ^= gfRandom << 5;	
}


void USART2_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t test;

    if (USART2->ISR & USART_ISR_RXNE)  {
		test = USART2->RDR & 0x00FF;
		if (test == CTRL_B) gfTermInput = CTRL_B;
		if (test != CTRL_D)  {
			*gUSART_RXHEAD = (uint8_t)test;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
		else  op_DEBUG();
    }
#ifdef USART_TX_INTERRUPT
    if (USART2->ISR & USART_ISR_TXE)  {
        if (gUSART_TXTAIL == gUSART_TXHEAD){ //nothing to transmit
            USART2->CR1 &= ~USART_CR1_TXEIE;  //then switch interrupt off
            return;
        }
        USART2->TDR = *gUSART_TXTAIL++ ;
        if (gUSART_TXTAIL == gUSART_TXBUFTOP)  gUSART_TXTAIL=gUSART_TXBUF;
    }
#endif
}

// #########################################################################
